from collections import defaultdict

import syntropy_sdk as sdk
from syntropy_sdk import models
from syntropy_sdk.utils import MAX_QUERY_FIELD_SIZE

from syntropynac import fields, transform
from syntropynac.fields import ConfigFields, PeerState, PeerType, Topology


def get_agents_connections(api, agents):
    ids = list(agents.keys())
    connections = (
        sdk.ConnectionsApi(api)
        .v1_network_connections_search(
            body=models.V1NetworkConnectionsSearchRequest(
                filter=models.V1ConnectionFilter(agent_id=ids),
            ),
        )
        .to_dict()["data"]
    )
    return connections


def export_connections(api, all_agents, network, net_agents, connections, topology):
    topology = Topology.P2M if not topology else topology.upper()
    ids = [connection["agent_connection_group_id"] for connection in connections]
    if ids:
        connections_services = sdk.utils.BatchedRequestFilter(
            sdk.ConnectionsApi(api).v1_network_connections_services_get,
            MAX_QUERY_FIELD_SIZE,
        )(filter=ids, _preload_content=False)["data"]

        connection_services = {
            connection["agent_connection_group_id"]: connection
            for connection in connections_services
        }
    net_connections = [
        {
            **connection,
            "agent_connection_services": connection_services.get(
                connection["agent_connection_group_id"], {}
            ),
        }
        for connection in connections
    ]
    transformed_connections = transform.transform_connections(
        all_agents,
        net_connections,
        topology if topology else network[fields.ConfigFields.TOPOLOGY],
    )
    if transformed_connections:
        network[fields.ConfigFields.CONNECTIONS] = transformed_connections
    network[fields.ConfigFields.TOPOLOGY] = topology

    # NOTE: Currently, SDN is disabled.
    if fields.ConfigFields.USE_SDN in network:
        del network[fields.ConfigFields.USE_SDN]

    # Filter out unused endpoints
    used_endpoints = [
        con[agent]["agent_id"]
        for con in net_connections
        for agent in ("agent_1", "agent_2")
    ]
    unused_endpoints = [id for id in net_agents if id not in used_endpoints]

    if unused_endpoints:
        agents_services = sdk.utils.BatchedRequestFilter(
            sdk.AgentsApi(api).v1_network_agents_services_get,
            MAX_QUERY_FIELD_SIZE,
        )(filter=unused_endpoints, _preload_content=False)["data"]

        agent_services = defaultdict(list)
        for agent_id, agent in zip(unused_endpoints, agents_services):
            id = (
                agent["agent_id"]
                if "agent_id" in agent and agent["agent_id"]
                else agent_id
            )
            agent_services[id].append(agent)

        network[fields.ConfigFields.ENDPOINTS] = {
            all_agents[id]["agent_name"]: {
                fields.ConfigFields.ID: id,
                fields.ConfigFields.SERVICES: [
                    service["agent_service_name"] for service in agent_services[id]
                ],
                fields.ConfigFields.TAGS: [
                    tag["agent_tag_name"]
                    for tag in all_agents[id].get("agent_tags", [])
                ],
            }
            for id in unused_endpoints
        }

    return network


def export_network(api, all_agents, topology):
    """Generate a network configuration structure from network and connections either
    using specified topology or inferred topology.
    Currently, default topology is P2M.

    Args:
        api (PlatformApi): Instance of PlatformApi.
        all_agents (dict[int, dict]): A mapping of all user agents ids to agent objects.
        network (dict): A dictionary describing a network to be exported.
        topology (str): One of MetadataNetworkType.

    Returns:
        dict: A network configuration structure.
    """
    net = {
        ConfigFields.TOPOLOGY: Topology.P2M,
        ConfigFields.STATE: PeerState.PRESENT,
    }

    connections = get_agents_connections(api, all_agents)

    net_agents = [agent["agent_id"] for _, agent in all_agents.items()]

    return export_connections(api, all_agents, net, net_agents, connections, topology)



#!/usr/bin/env python
import json

import click
import syntropy_sdk as sdk
import yaml

from syntropynac import configure as configure_module
from syntropynac import fields, transform, utils
from syntropynac.decorators import syntropy_api


@click.group()
def apis():
    """Syntropy Network As Code Command Line Interface."""


@apis.command()
@click.argument("config")
@click.option(
    "--dry-run",
    is_flag=True,
    default=False,
    help="Perform a dry run without configuring anything.",
)
@click.option(
    "--json",
    "-j",
    "from_json",
    is_flag=True,
    default=False,
    help="Imports configuration from JSON instead of YAML.",
)
@syntropy_api
def configure(config, dry_run, from_json, api):
    """Configure connections using a configuration YAML/JSON file.

    \b
    Example YAML file:
        name: test-network
        state: present
        topology: P2M
        connections:
            gateway-endpoint:
                state: present
                type: endpoint
                services:
                - postgres
                - redis
                connect_to:
                    endpoint-1:
                        type: endpoint
                        services:
                        - app
                    endpoint2:
                        state: present
                        type: endpoint
                        services:
                        - app
    """
    try:
        with open(config, "rb") as cfg_file:
            if from_json:
                config = json.load(cfg_file)
                config = config if isinstance(config, list) else [config]
            else:
                config = list(yaml.safe_load_all(cfg_file))
    except FileNotFoundError:
        click.secho(f"Could not find {config} file.", err=True, fg="red")
        return
    except json.decoder.JSONDecodeError:
        click.secho(f"Could not parse {config} file as JSON.", err=True, fg="red")
        return
    except yaml.YAMLError:
        click.secho(f"Could not parse {config} file as YAML.", err=True, fg="red")
        return

    for index, net in enumerate(config):
        if any(i not in net for i in ("topology", "state")):
            click.secho(
                f"Skipping {index} entry as no name, topology or state found.",
                fg="yellow",
            )
            continue
        configure_module.configure_network(api, net, dry_run)

    click.secho("Done", fg="green")


@apis.command()
@click.option("--topology", default=None, type=str, help="Override network topology.")
@click.option(
    "--json",
    "-j",
    "to_json",
    is_flag=True,
    default=False,
    help="Outputs a JSON instead of YAML.",
)
@syntropy_api
def export(topology, to_json, api):
    """Exports existing connections to configuration YAML/JSON file.

    If exact topology export is required - use P2P topology.
    """
    all_agents = sdk.utils.WithPagination(sdk.AgentsApi(api).v1_network_agents_get)(
        _preload_content=False
    )["data"]
    all_agents = {agent["agent_id"]: agent for agent in all_agents}

    network = utils.export_network(api, all_agents, topology)
    if to_json:
        click.echo(json.dumps(network, indent=4))
    else:
        click.echo(yaml.dump_all([network]))


def main():
    apis(prog_name="syntropynac")


if __name__ == "__main__":
    main()



import click
import syntropy_sdk as sdk
from syntropy_sdk import models, utils

from syntropynac import resolve, transform
from syntropynac.exceptions import ConfigureNetworkError
from syntropynac.fields import ALLOWED_TOPOLOGIES, ConfigFields, PeerState, Topology


def create_connections(api, peers, silent=False):
    body = models.V1NetworkConnectionsCreateP2PRequest(
        agent_pairs=[
            models.V1NetworkConnectionsCreateP2PRequestAgentPairs(
                agent_1_id=a,
                agent_2_id=b,
            )
            for a, b in peers
        ],
    )
    sdk.ConnectionsApi(api).v1_network_connections_create_p2_p(
        body=body, _preload_content=False
    )

    connections = utils.WithPagination(
        sdk.ConnectionsApi(api).v1_network_connections_get
    )(_preload_content=False)["data"]

    frozen_peers = [frozenset(peer) for peer in peers]
    connections = [
        con
        for con in connections
        if frozenset((con["agent_1"]["agent_id"], con["agent_2"]["agent_id"]))
        in frozen_peers
    ]

    not silent and click.echo(f"Created {len(connections)} connections")

    return connections


def delete_connections(api, absent):
    connections = (
        sdk.ConnectionsApi(api)
        .v1_network_connections_search(
            body=models.V1NetworkConnectionsSearchRequest(
                filter=models.V1ConnectionFilter(
                    agent_pair=[
                        models.V1AgentPairFilter(agent_1_id=a, agent_2_id=b)
                        for a, b in absent
                    ]
                )
            ),
        )
        .to_dict()["data"]
    )

    sdk.ConnectionsApi(api).v1_network_connections_remove(
        body=models.V1NetworkConnectionsRemoveRequest(
            agent_connection_group_ids=[
                conn["agent_connection_group_id"] for conn in connections
            ],
        ),
    )


def configure_connection(api, config, connection, silent=False):
    agents = {
        connection["agent_1"]["agent_id"]: connection["agent_1"],
        connection["agent_2"]["agent_id"]: connection["agent_2"],
    }

    enabled_subnets = set(config.get_subnets(1, agents) + config.get_subnets(2, agents))
    current_subnets = {
        subnet["agent_service_subnet_id"]: subnet["agent_connection_subnet_is_enabled"]
        for subnet in connection["agent_connection_subnets"]
    }

    # First collect all the changes to the original configured subnets
    changes = [
        models.AgentServicesUpdateChanges(
            agent_service_subnet_id=id,
            is_enabled=id in enabled_subnets,
        )
        for id, enabled in current_subnets.items()
        if (id in enabled_subnets) != enabled
    ]
    # Then configure any missing subnets
    changes += [
        models.AgentServicesUpdateChanges(
            agent_service_subnet_id=id,
            is_enabled=True,
        )
        for id in enabled_subnets
        if id not in current_subnets
    ]

    if not changes:
        return 0

    body = models.V1NetworkConnectionsServicesUpdateRequest(
        agent_connection_group_id=connection["agent_connection_group_id"],
        changes=changes,
    )
    sdk.ConnectionsApi(api).v1_network_connections_services_update(body=body)
    return len(changes)


def configure_connections(api, services_config, connections, silent=False):
    ids = [connection["agent_connection_group_id"] for connection in connections]
    if not ids:
        return 0, 0
    connections_services = utils.BatchedRequestFilter(
        sdk.ConnectionsApi(api).v1_network_connections_services_get,
        utils.MAX_QUERY_FIELD_SIZE,
    )(filter=ids, _preload_content=False)["data"]

    # Build a map of connections so that it would be quicker to resolve them to subnets
    services_map = {}
    for conn in connections_services:
        services_map[
            frozenset((conn["agent_1"]["agent_id"], conn["agent_2"]["agent_id"]))
        ] = conn

    updated_connections = 0
    updated_subnets = 0
    # Update subnets with connection subnets
    for config in services_config:
        key = frozenset((config.agent_1, config.agent_2))
        if key not in services_map:
            not silent and click.secho(
                f"Warning: Connection from {config.agent_1} to {config.agent_2} was not created.",
                fg="yellow",
                err=True,
            )
            continue

        updated_connections += 1
        updated_subnets += configure_connection(
            api, config, services_map[key], silent=silent
        )

    return updated_connections, updated_subnets


def configure_network_update(api, config, dry_run, silent=False):
    """Updates existing network's connection.
    NOTE: This will ignore any preconfigured connections that are not
    explicitly specified in the config dictionary.

    Example config dictionary:
    {
        "topology": `P2P|P2M|MESH`,
        "use_sdn": `True|False`,
        "state": "present|absent",
        "connections": {
            # Connection for P2P
            "endpoint-name-1": {
                "type": "endpoint|tag|id",
                "id": 1,
                "state": "present|absent",
                "services": ["service1", "service2"],
                "connect_to": {
                    "endpoint-name-2": {
                        "type": "endpoint",
                        "id": 2,
                        "services": ["service3", "service4"],
                    },
                },
            },

            # Connection for P2M
            "source-endpoint": {
                "type": "endpoint",
                "id": 1,
                "state": "present",
                "connect_to": {
                    "destination-1": {
                        "state": "present",
                        "id": 10,
                        "type": "endpoint",
                    },
                    "destination-2": {
                        "state": "present",
                        "id": 11,
                        "type": "endpoint",
                    },
                    "destination-3": {
                        "state": "present",
                        "id": 12,
                        "type": "endpoint",
                    },
                },
            },

            # MESH network - all endpoints interconnected
            "endpoint-1": {
                "type": "endpoint",
                "state": "present",
                "id": 10,
            },
            "endpoint-2": {
                "type": "endpoint",
                "state": "present",
                "id": 11,
            },
            "endpoint-3": {
                "type": "endpoint",
                "state": "present",
                "id": 12,
            },
            "endpoint-4": {
                "type": "endpoint",
                "state": "present",
                "id": 13,
            },
        }
    }

    Args:
        api (PlatformApi): Instance of the platform API.
        config (dict): Configuration dictionary.
        dry_run (bool): Indicates whether to perform a dry run (without any configuration).
        silent (bool, optional): Indicates whether to suppress messages - used with Ansible. Defaults to False.
    Returns:
        (bool): True if any changes were made and False otherwise
    """
    topology = config[ConfigFields.TOPOLOGY].upper()
    connections = utils.WithPagination(
        sdk.ConnectionsApi(api).v1_network_connections_get
    )(_preload_content=False)["data"]
    all_agents = resolve.get_all_agents(api, silent)
    resolved_connections = transform.transform_connections(
        all_agents,
        connections,
        topology,
        group_tags=False,
        silent=silent,
    )
    current_connections = {
        frozenset(
            (connection["agent_1"]["agent_id"], connection["agent_2"]["agent_id"])
        ): connection
        for connection in connections
    }
    config_connections = config.get(ConfigFields.CONNECTIONS, {})

    if topology == Topology.P2P:
        present, absent, services = resolve.resolve_p2p_connections(
            api, config_connections, silent=silent
        )
    elif topology == Topology.P2M:
        present, absent, services = resolve.resolve_p2m_connections(
            api, config_connections, silent=silent
        )
    else:
        present, absent, services = resolve.resolve_mesh_connections(
            api, config_connections, silent=silent
        )
    if topology == Topology.P2P:
        current, _, _ = resolve.resolve_p2p_connections(
            api, resolved_connections, silent=silent
        )
    elif topology == Topology.P2M:
        current, _, _ = resolve.resolve_p2m_connections(
            api, resolved_connections, silent=silent
        )
    else:
        current, _, _ = resolve.resolve_mesh_connections(
            api, resolved_connections, silent=silent
        )

    present = [frozenset(i) for i in present]
    absent = [frozenset(i) for i in absent]
    current = [frozenset(i) for i in current]

    to_add = [list(link) for link in present if link not in current]

    if dry_run:
        not silent and click.echo(f"Would remove {len(absent)} connections.")
    else:
        delete_connections(api, absent)
        not silent and click.echo(f"Removed {len(absent)} connections.")

    added_connections = []
    if dry_run:
        not silent and click.echo(f"Would create {len(to_add)} connections.")
    elif to_add:
        added_connections = create_connections(api, to_add, silent)

    to_remove = [
        conn["agent_connection_group_id"]
        for link, conn in current_connections.items()
        if link in absent
    ]
    connections = [
        connection
        for connection in connections + added_connections
        if connection["agent_connection_group_id"] not in to_remove
    ]

    if dry_run:
        not silent and click.echo(f"Would configure {len(connections)} connections.")
    else:
        updated_connections, updated_subnets = configure_connections(
            api, services, connections, silent=silent
        )
        not silent and click.echo(
            f"Configured {updated_connections} connections and {updated_subnets} subnets"
        )
        return True
    return False


def configure_network_delete(api, config, dry_run, silent=False):
    """Deletes existing network's connections and the network itself.

    Args:
        api (PlatformApi): Instance of the platform API.
        network (dict): Dictionary containing id and name keys.
        dry_run (bool): Indicates whether to perform a dry run (without any configuration).
        silent (bool, optional): Indicates whether to suppress messages - used with Ansible. Defaults to False.

    Returns:
        (bool): True if any changes were made and False otherwise
    """
    config_connections = config.get(ConfigFields.CONNECTIONS, {})
    topology = config[ConfigFields.TOPOLOGY].upper()

    if topology == Topology.P2P:
        _, absent, _ = resolve.resolve_p2p_connections(
            api, config_connections, silent=silent
        )
    elif topology == Topology.P2M:
        _, absent, _ = resolve.resolve_p2m_connections(
            api, config_connections, silent=silent
        )
    else:
        _, absent, _ = resolve.resolve_mesh_connections(
            api, config_connections, silent=silent
        )

    if dry_run:
        not silent and click.echo(f"Would delete {len(absent)} connections...")
        return False
    else:
        delete_connections(api, absent)
        return True


def configure_network(api, config, dry_run, silent=False):
    """Configures Syntropy Network based on the current state and the requested state.

    Args:
        api (PlatformApi): Instance of the platform API.
        config (dict): Configuration dictionary.
        dry_run (bool): Indicates whether to perform a dry run (without any configuration).
        silent (bool, optional): Indicates whether to suppress messages - used with Ansible. Defaults to False.

    Returns:
        (bool): True if any changes were made and False otherwise
    """
    if not all(i in config for i in (ConfigFields.TOPOLOGY, ConfigFields.STATE)):
        error = f"{ConfigFields.TOPOLOGY} and {ConfigFields.STATE} must be present"
        if not silent:
            click.secho(error, err=True, fg="red")
        else:
            raise ConfigureNetworkError(error)
        return False

    state = config[ConfigFields.STATE]
    if state not in (PeerState.PRESENT, PeerState.ABSENT):
        error = f"Invalid state {state}"
        if not silent:
            click.secho(error, fg="red", err=True)
            return False
        else:
            raise ConfigureNetworkError(error)

    if not resolve.validate_connections(
        config.get(ConfigFields.CONNECTIONS, {}), silent
    ):
        error = f"Invalid {ConfigFields.CONNECTIONS} format."
        if not silent:
            click.secho(error, fg="red", err=True)
            return False
        else:
            raise ConfigureNetworkError(error)

    not silent and click.secho(f"Configuring network", fg="green")

    if state == PeerState.PRESENT:
        return configure_network_update(api, config, dry_run, silent=silent)
    elif state == PeerState.ABSENT:
        return configure_network_delete(api, config, dry_run, silent=silent)
    return False
